<!DOCTYPE kpartgui>
<kpartgui name="ksysv" version="3">
  <MenuBar noMerge="1">
	<Menu name="file"><text>&amp;File</text>
	  <Action name="file_revert"/>
	  <Action name="file_open"/>
	  <Separator line_separator="true"/>
	  <Action name="file_save"/>
	  <Action name="file_save_as"/>
	  <Action name="ksysv_save_log"/>
	  <Separator line_separator="true"/>
	  <Action name="ksysv_print_log"/>
	  <Separator line_separator="true"/>
	  <Action name="file_quit"/>
	</Menu>

	<Menu name="edit"><text>&amp;Edit</text>
	  <Action name="edit_undo"/>
	  <Action name="edit_redo"/>
	  <Separator line_separator="true"/>
	  <Action name="edit_cut"/>
	  <Action name="edit_copy"/>
	  <Action name="edit_paste"/>
	  <Separator line_separator="true"/>
	  <Action name="ksysv_properties"/>
	</Menu>
	
	<Menu name="tools"><text>&amp;Tools</text>
	  <Action name="ksysv_start_service"/>
	  <Action name="ksysv_stop_service"/>
	  <Action name="ksysv_restart_service"/>
	  <Separator line_separator="true"/>
	  <Action name="ksysv_edit_service"/>
	</Menu>

	<Menu name="settings"><text>&amp;Settings</text>
          <Merge name="StandardToolBarMenuHandler" />
	  <Action name="ksysv_toggle_log"/>
	  <Separator line_separator="true"/>
	  <Action name="options_save_options"/>
	  <Separator line_separator="true"/>
	  <Action name="options_configure_keybinding"/>
	  <Action name="options_configure_toolbars"/>
	  <Action name="options_configure"/>
	</Menu>

	<Menu name="help"><text>&amp;Help</text>
      <Action name="help_contents"/>
	  <Action name="help_whats_this"/>
	  <Separator line_separator="true"/>
	  <Action name="help_report_bug"/>
	  <Separator line_separator="true"/>                                   
	  <Action name="help_about_app"/>
	  <Action name="help_about_kde"/>
	</Menu>
  </MenuBar>
  
  <ToolBar fullWidth="true" name="mainToolBar" noMerge="1">
	<text>Main Toolbar</text>
	<Action name="file_revert"/>
	<Action name="file_save"/>
	<Separator line_separator="true"/>
	<Action name="edit_undo"/>
	<Action name="edit_redo"/>
	<Separator line_separator="true"/>
	<Action name="edit_cut"/>
	<Action name="edit_copy"/>
	<Action name="edit_paste"/>
	<Action name="ksysv_toggle_log"/>
  </ToolBar>
  
  <Menu name="item_menu">
	<text>Runlevel Menu</text>
	<Action name="edit_cut"/>
	<Action name="edit_copy"/>
	<Action name="edit_paste"/>
	<Separator/>
	<Action name="ksysv_properties"/>
  </Menu>
  
  <Menu name="script_menu">
	<text>Services Menu</text>
	<Action name="edit_copy"/>
	<Separator/>
	<Action name="ksysv_open_service"/>
	<Action name="ksysv_open_with"/>
	<Separator/>
	<Action name="ksysv_properties"/>
  </Menu>
  
  <Menu name="list_menu">
	<text>Runlevel Menu</text>
	<Action name="ksysv_paste_append"/>
  </Menu>
  
</kpartgui>
